/*!
  \file
  \brief LIDAR との接続を開く

  \author Satofumi Kamimura
*/

#include <iostream>
#include "open_lidar_device.h"
#include "Urg_driver.h"
#include "Lms_driver.h"
#include "log_printf.h"

using namespace hrk;
using namespace std;


namespace
{
    typedef enum { Urg, Lms } lidar_t;


    bool set_connection(string& device_or_address, long& baudrate_or_port,
                        const lidar_t& sensor, const Lidar::connection_t& type)
    {
        if (type == Lidar::Serial) {
            if (sensor == Urg) {
                vector<string> devices = Urg_driver::find_ports();
                if (devices.empty()) {
                    cerr << "can not find a serial devide." << endl;
                    return false;
                }
                device_or_address = devices.front();
                baudrate_or_port = Urg_driver::Default_baudrate;

            } else {
                // Lms
                // !!!
            }

        } else {
            // Ethernet
            if (sensor == Urg) {
                device_or_address = "192.168.0.10";
                baudrate_or_port = Urg_driver::Default_port;

            } else {
                // Lms
                device_or_address = "192.168.0.1";
                baudrate_or_port = Lms_driver::Default_port;
            }
        }

        return true;
    }


    Lidar* open_connection(const lidar_t& sensor,
                           const string& device_or_address,
                           long baudrate_or_port,
                           const Lidar::connection_t& type)
    {
        Lidar* lidar;

        Urg_driver* urg;
        Lms_driver* lms;
        bool is_open = false;
        switch (sensor) {
        case Urg:
            urg = new Urg_driver;
            is_open = urg->open(device_or_address, baudrate_or_port, type);
            lidar = urg;
            break;

        case Lms:
            lms = new Lms_driver;
            is_open = lms->open(device_or_address, baudrate_or_port, type);
            lidar = lms;
            break;
        }

        if (!is_open) {
            cerr << "can not open: " << device_or_address << ", "
                 << baudrate_or_port << endl;
        }

    return lidar;
    }
}


hrk::Lidar* hrk::open_lidar_device(int argc, char *argv[])
{
    lidar_t sensor = Urg;
    Lidar::connection_t type = Lidar::Serial;

    for (int i = 1; i < argc; ++i) {
        const string token = argv[i];

        if (token == "-lms") {
            sensor = Lms;

        } else if (token == "-e") {
            type = Lidar::Ethernet;
        }
    }

    string device_or_address;
    long baudrate_or_port;
    if (!set_connection(device_or_address, baudrate_or_port, sensor, type)) {
        return new Urg_driver;
    }

    return open_connection(sensor, device_or_address, baudrate_or_port, type);
}
