#ifndef HRK_LIDAR_H
#define HRK_LIDAR_H

/*!
  \file
  \brief Lidar インターフェース

  \author Satofumi Kamimura
*/

#include <vector>
#include <string>


namespace hrk
{
    class Connection;

    //! Lidar インターフェース
    class Lidar
    {
    public:
        enum {
            Infinity_scan_times = -1,
        };

        typedef enum {
            Serial,
            Ethernet,
        } connection_t;

        typedef enum {
            Distance,            //!< 距離
            Distance_intensity,  //!< 距離 + 強度
            Multiecho,           //!< マルチエコーの距離
            Multiecho_intensity, //!< マルチエコーの(距離 + 強度)
        } measurement_t;

        virtual ~Lidar()
        {
        }

        /*!
          \brief 状態を示すメッセージを返す

          \return メッセージ文字列
        */
        virtual const char* what() const = 0;

        virtual bool open(const std::string& device_name_or_ip_address,
                          long baudrate_or_port_number,
                          connection_t type = Serial) = 0;
        virtual bool open(Connection* connection) = 0;

        //! 接続を閉じる
        virtual void close() = 0;

        //! センサが起動中かどうかを返す
        virtual bool is_booting() = 0;

        //! 接続しているかを返す
        virtual bool is_open() const = 0;

        //! 接続オブジェクトを設定する
        virtual void set_connection(Connection* connection) = 0;

        //! 接続オブジェクトを返す
        virtual Connection* connection() = 0;

        virtual bool start_measurement(measurement_t type = Distance,
                                       int scan_times = Infinity_scan_times,
                                       int skip_scan = 0) = 0;
        virtual bool get_distance(std::vector<long>& data,
                                  long *time_stamp) = 0;
        virtual bool get_distance_intensity(std::vector<long>& data,
                                            std::vector<unsigned short>&
                                            intensity,
                                            long *time_stamp) = 0;
        virtual bool get_multiecho(std::vector<long>& data_multi,
                                   long* time_stamp) = 0;
        virtual bool get_multiecho_intensity(std::vector<long>& data_multiecho,
                                             std::vector<unsigned short>&
                                             intensity_multiecho,
                                             long* time_stamp) = 0;
        virtual bool set_scanning_parameter(int first_step, int last_step,
                                            int skip_step) = 0;
        virtual void stop_measurement() = 0;

        virtual double index2rad(int index) const = 0;
        virtual double index2deg(int index) const = 0;
        virtual int rad2index(double radian) const = 0;
        virtual int deg2index(double degree) const = 0;

        virtual double step2rad(int step) const = 0;
        virtual double step2deg(int step) const = 0;
        virtual int rad2step(double radian) const = 0;
        virtual int deg2step(double degree) const = 0;

        virtual int min_step() const = 0;
        virtual int max_step() const = 0;
        virtual int front_step() const = 0;
        virtual int total_steps() const = 0;
        virtual long min_distance() const = 0;
        virtual long max_distance() const = 0;
        virtual long scan_usec() const = 0;
        virtual int max_data_size() const = 0;
        virtual int max_echo_size() const = 0;

        virtual std::string sensor_series_name() const = 0;
        virtual std::string sensor_product_name() const = 0;
        virtual std::string sensor_product_version() const = 0;
        virtual std::string serial_id() const = 0;
    };
}

#endif /* !HRK_LIDAR_H */
