/*!
  \file
  \brief LMS Sensor

  \author Satofumi Kamimura
*/

#include <cstdlib>
#include <sstream>
#include "Lms_driver.h"
#include "Serial.h"
#include "Tcpip.h"
#include "connection_utils.h"
#include "token_utils.h"
#include "math_utils.h"

using namespace hrk;
using namespace std;


namespace
{
    const char STX = '\02';
    const char ETX = '\03';

    enum {
        Hex = 16,

        Default_timeout = 1000,
    };

    const int Scan_range = 190;
}


struct Lms_driver::pImpl
{
    string error_message_;
    Connection* created_connection_;
    Connection* connection_;

    long timeout_;
    long scan_usec_;
    int total_steps_;
    int min_index_;
    int max_index_;
    int front_index_;
    int start_index_;
    int stop_index_;
    bool is_multiecho_;

    measurement_t measurement_type_;
    int left_scan_times_;
    string sensor_series_name_;
    string sensor_product_name_;
    string sensor_product_version_;
    string serial_id_;


    pImpl() : error_message_("no error."),
              created_connection_(NULL), connection_(NULL),
              timeout_(Default_timeout), scan_usec_(0), total_steps_(0),
              min_index_(-1), max_index_(-1),
              front_index_(-1), start_index_(-1), stop_index_(-1),
              is_multiecho_(false),
              measurement_type_(Distance), left_scan_times_(0),
              sensor_series_name_("Unknown"), sensor_product_name_("Unknown"),
              sensor_product_version_("Unknown"), serial_id_("Unknown")
    {
    }


    bool open(const std::string& device_name_or_ip_address,
              long baudrate_or_port_number, connection_t type)
    {
        if (!create_connection(device_name_or_ip_address,
                               baudrate_or_port_number, type)) {
            close();
            return false;
        }

        if (!read_parameters()) {
            close();
            return false;
        }
        connection_ = created_connection_;

        return true;
    }


    bool open(Connection* connection)
    {
        created_connection_ = connection;
        if (!read_parameters()) {
            created_connection_ = NULL;
            close();
            return false;
        }
        connection_ = created_connection_;
        created_connection_ = NULL;

        return true;
    }


    bool is_open() const
    {
        return (connection_) ? connection_->is_open() : false;
    }


    void close()
    {
        if (connection_) {
            connection_->close();
        }
        if (created_connection_) {
            delete created_connection_;
            created_connection_ = NULL;
        }
        connection_ = NULL;

        sensor_series_name_.clear();
        sensor_product_name_.clear();
        sensor_product_version_.clear();
    }


    bool create_connection(const string& device_name_or_ip_address,
                           long baudrate_or_port_number, connection_t type)
    {
        if (type == Serial) {
            hrk::Serial* serial_connection = new hrk::Serial;
            if (!serial_connection->open(device_name_or_ip_address,
                                         baudrate_or_port_number)) {
                error_message_ = serial_connection->what();
                return false;
            }
            created_connection_ = serial_connection;

        } else if (type == Ethernet) {
            Tcpip* tcpip_connection = new Tcpip;
            if (!tcpip_connection->open(device_name_or_ip_address,
                                        baudrate_or_port_number)) {
                error_message_ = tcpip_connection->what();
                return false;
            }
            created_connection_ = tcpip_connection;

        } else {
            error_message_ = "unknown connection type.";
            return false;
        }

        return true;
    }


    bool lms_send(Connection* connection, const string& raw_command)
    {
        const string command = STX + raw_command + ETX;
        int send_size = command.size();
        int n = connection->write(command.c_str(), send_size);
        if (n != send_size) {
            return false;
        }

        return true;
    }


    string lms_response(Connection* connection,
                        const string& raw_command, long timeout)
    {
        if (!lms_send(connection, raw_command)) {
            return "";
        }

        enum { Buffer_size = 256 };
        char buffer[Buffer_size + 1];
        int n = read_to_ETX(connection, buffer, Buffer_size, timeout);
        if (n < 2) {
            return "";
        }
        if (buffer[n - 1] == ETX) {
            --n;
        }
        buffer[n] = '\0';

        return &buffer[1];
    }


    int read_to_ETX(Connection* connection, char buffer[],
                    int buffer_size, long timeout)
    {
        return readline(connection, buffer, buffer_size, timeout, ETX);
    }


    bool read_parameters()
    {
        Connection* connection = created_connection_;
        string response = lms_response(connection, "sRN LMPscancfg", timeout_);
        if (!parse_frequency_and_angular_resolution(response)) {
            return false;
        }

        response = lms_response(connection, "sRN LMDscandata", timeout_);
        ignore_to(connection, timeout_, ETX);
        if (!parse_serial_id(response)) {
            return false;
        }

        response = lms_response(connection, "sRN DeviceIdent", timeout_);
        if (!parse_device_ident(response)) {
            return false;
        }

        return true;
    }


    long parse_Uint_32(const string& text_value, int base = Hex)
    {
        return strtol(text_value.c_str(), NULL, base);
    }


    long parse_Int_32(const string& text_value, int base = Hex)
    {
        long value = parse_Uint_32(text_value, base);
        if (static_cast<unsigned long>(value) >= 0x80000000) {
            value = -((~value + 1) & 0xffffffff);
        }
        return value;
    }


    long Hz_to_Usec(int hz)
    {
        return 1000 * 1000 / hz;
    }


    bool parse_frequency_and_angular_resolution(const string& response)
    {
        if (response.empty()) {
            return false;
        }

        enum {
            Token_size = 7,
            Frequency_index = 2,
            Angle_resolution_index = 4,
            Start_angle_index = 5,
            Stop_angle_index = 6,
        };
        const vector<string> tokens = split(response, " ");
        if (tokens.size() != Token_size) {
            return false;
        }

        //! \todo LMS1xx の場合の処理を追加する

        // Scan Frequency [1/100 Hz]
        long scan_usec = 0;
        long value = parse_Uint_32(tokens[Frequency_index]);
        switch (value) {
        case 0x9C4:
            scan_usec = Hz_to_Usec(25);
            break;

        case 0xDAC:
            scan_usec = Hz_to_Usec(35);
            break;

        case 0x1388:
            scan_usec = Hz_to_Usec(50);
            break;

        case 0x1A0B:
            scan_usec = Hz_to_Usec(75);
            break;

        case 0x2710:
            scan_usec = Hz_to_Usec(100);
            break;

        default:
            return false;
        }

        // Angle Resolution [1/10000 deg]
        int total_steps = 0;
        value = parse_Uint_32(tokens[Angle_resolution_index]);
        switch (value) {
        case 0x683:
            total_steps = 360.0 / 0.166667;
            break;

        case 0x9C4:
            total_steps = 360.0 / 0.25;
            break;

        case 0xD05:
            total_steps = 360.0 / 0.333333;
            break;

        case 0x1388:
            total_steps = 360.0 / 0.5;
            break;

        case 0x1A0B:
            total_steps = 360.0 / 0.666667;
            break;

        case 0x2710:
            total_steps = 360.0 / 1.0;
            break;

        default:
            return false;
        }

        // Start Angle [1/10000 deg]
        int min_index = 0;
        long start_angle = parse_Int_32(tokens[Start_angle_index]);

        // Stop Angle [1/10000 deg]
        long stop_angle = parse_Int_32(tokens[Stop_angle_index]);
        int max_index = Scan_range / (360.0 / total_steps);

        scan_usec_ = scan_usec;
        total_steps_ = total_steps;
        min_index_ = min_index;
        max_index_ = max_index;
        front_index_ = (max_index - min_index) / 2;

        start_index_ = deg2index((start_angle / 10000.0) - 90);
        stop_index_ = deg2index((stop_angle / 10000.0) - 90);
        return true;
    }


    bool parse_serial_id(const string& response)
    {

        enum { Serial_id_index = 4 };
        const vector<string> tokens = split(response, " ");
        if (static_cast<int>(tokens.size()) <= Serial_id_index) {
            return false;
        }

        serial_id_ = tokens[Serial_id_index];
        return true;
    }


    int deg2index(double degree) const
    {
        return rad2index(degree * M_PI / 180.0);
    }


    int rad2index(double radian) const
    {
        int step = rad2step(radian);

        //! \todo multiecho に対応させる
        if (is_multiecho_) {
            step *= max_echo_size();
        }
        return step;
    }


    int max_echo_size() const
    {
        return 5;
    }


    bool parse_device_ident(const string& response)
    {
        enum {
            Token_size = 6,
            Product_name_index = 3,
            Product_version_index = 5,
            Serial_id_index = 4,
        };
        vector<string> tokens = split(response, " ");
        if (tokens.size() != Token_size) {
            return false;
        }

        sensor_series_name_ = "LMS";
        sensor_product_name_ = tokens[Product_name_index];
        vector<string> product_tokens = split(sensor_product_name_, "_");
        if (!product_tokens.empty()) {
            sensor_product_name_ = product_tokens[0];
        }
        sensor_product_version_ = tokens[Product_version_index];
        serial_id_ = tokens[Serial_id_index];

        return true;
    }


    int rad2step(double radian) const
    {
        int step = static_cast<int>(floor(total_steps_ *
                                          radian / (2.0 * M_PI)) + 0.5) +
            front_index_ - min_index_;
        return min(max(0, step), max_index_);
    }


    bool receive_data(long data[], unsigned short intensity[], long *timestamp)
    {
        static_cast<void>(intensity);

        enum {
            Scale_factor_index = 21,
            Data_1st_index = 26,
            Time_index = 9,
        };

        enum { Buffer_size = 8196 + 2048 };
        char buffer[Buffer_size];
        int n = read_to_ETX(connection_, buffer, Buffer_size, timeout_);
        //fprintf(stderr, "n = %d\n", n);
        if (n <= 0) {
            return false;
        }
        buffer[n] = '\0';
        //fprintf(stderr, "[%s]\n", buffer);

        const vector<string> tokens = split(buffer, " ");
        //fprintf(stderr, "tokens: %ld\n", tokens.size());
        if (static_cast<int>(tokens.size()) < Data_1st_index) {
            return false;
        }

        // 距離データの格納
        int scale_factor = (tokens[Scale_factor_index] == "40000000") ? 2 : 1;
        const int echoes = is_multiecho_ ? max_echo_size() : 1;
        for (int k = 0; k < echoes; ++k) {
            enum { Footer_token_size = 6 + 1 };
            int echo_index = k * (max_index_ + Footer_token_size);
            //fprintf(stderr, "<%d>, ", echo_index);
            for (int i = 0; i <= max_index_; ++i) {
                int index = Data_1st_index + echo_index + i;
                if (index >= static_cast<int>(tokens.size())) {
                    break;
                }
                //fprintf(stderr, "%d, ", echo_index + i);
                data[(echoes * i) + k] =
                    scale_factor * parse_Uint_32(tokens[index], Hex);
            }
        }

        if (intensity) {
            // 強度データの格納
            enum { Footer_token_size = 6 + 1 };
            const int Intensity_1st_index = Data_1st_index + 1
                + (echoes * (max_index_ + Footer_token_size));
            for (int k = 0; k < echoes; ++k) {
                int echo_index = k * (max_index_ + Footer_token_size);
                for (int i = 0; i <= max_index_; ++i) {
                    int index = Intensity_1st_index + echo_index + i;
                    if (index >= static_cast<int>(tokens.size())) {
                        break;
                    }
                    intensity[(echoes * i) + k] =
                        parse_Uint_32(tokens[index], Hex) << 8;
                }
            }
        }

        if (timestamp) {
            *timestamp = parse_Uint_32(tokens[Time_index], Hex) / 1000;
        }

        if (left_scan_times_ > 0) {
            --left_scan_times_;
            if (left_scan_times_ <= 0) {
                stop_measurement();
            }
        }

        return true;
    }


    bool set_measurement(measurement_t type, int skip_scan)
    {
        bool is_intensity = false;
        bool is_multiecho = false;

        switch (type) {
        case Distance:
            break;

        case Distance_intensity:
            is_intensity = true;
            break;

        case Multiecho:
            is_multiecho = true;
            break;

        case Multiecho_intensity:
            is_multiecho = true;
            is_intensity = true;
            break;
        }

        set_access_mode("03");

        ostringstream stream;
        enum { Max_output_rate = 5000 };
        stream << min(max(0, skip_scan + 1), static_cast<int>(Max_output_rate));
        const string scan_skip_text = stream.str();

        const string command =
            string("sWN LMDscandatacfg 01 00 ") +
            (is_intensity ? "1 " : "0 ") + "0 0 00 00 0 0 0 0 +" +
            scan_skip_text;
        //fprintf(stderr, "[%s]\n", command.c_str());
        string response = lms_response(connection_, command, timeout_);
        //fprintf(stderr, "cfg response: %s\n", response.c_str());
        if (response != "sWA LMDscandatacfg") {
            return false;
        }

        if (sensor_product_name_ == "LMS5xx") {
            const string command = is_multiecho ?
                "sWN FREchoFilter 1" : "sWN FREchoFilter 0";
            response = lms_response(connection_, command, timeout_);
            //fprintf(stderr, "response: %s\n", response.c_str());
            if (response != "sWA FREchoFilter") {
                return false;
            }
        } else {
            // LMS1xx
            // !!!
        }

        is_multiecho_ = is_multiecho;

        set_to_run();
        return true;
    }


    bool set_access_mode(const string& mode)
    {
        const string response =
            lms_response(connection_, "sMN SetAccessMode " + mode + " F4724744",
                         timeout_);
        if (response != "sAN SetAccessMode 1") {
            return false;
        }

        return true;
    }


    bool set_to_run()
    {
        const string response = lms_response(connection_, "sMN Run", timeout_);
        if (response != "sAN Run 1") {
            return false;
        }

        return true;
    }


    bool start_measurement(int scan_times)
    {
        if (scan_times == 1) {
            return lms_send(connection_, "sRN LMDscandata");
        }

        const string response =
            lms_response(connection_, "sEN LMDscandata 1", timeout_);
        return (response == "sEA LMDscandata 1") ? true : false;
    }


    void stop_measurement()
    {
        lms_response(connection_, "sEN LMDscandata 0", timeout_);
    }
};


Lms_driver::Lms_driver() : pimpl(new pImpl)
{
}


Lms_driver::~Lms_driver()
{
}


const char* Lms_driver::what() const
{
    return pimpl->error_message_.c_str();
}


bool Lms_driver::open(const std::string& device_name_or_ip_address,
                      long baudrate_or_port_number, connection_t type)
{
    close();

    return pimpl->open(device_name_or_ip_address,
                       baudrate_or_port_number, type);
}


bool Lms_driver::open(Connection* connection)
{
    //close();
    return pimpl->open(connection);
}


bool Lms_driver::is_booting()
{
    // !!!
    return false;
}


void Lms_driver::close()
{
    pimpl->close();
}


bool Lms_driver::is_open() const
{
    return pimpl->is_open();
}


void Lms_driver::set_connection(Connection* connection)
{
    close();
    pimpl->connection_ = connection;
}


Connection* Lms_driver::connection()
{
    return pimpl->connection_;
}


bool Lms_driver::start_measurement(measurement_t type,
                                   int scan_times, int skip_scan)
{
    pimpl->set_measurement(type, skip_scan);
    pimpl->start_measurement(scan_times);

    pimpl->left_scan_times_ = scan_times;
    pimpl->measurement_type_ = type;
    return true;
}


bool Lms_driver::get_distance(std::vector<long>& data, long *time_stamp)
{
    if (!is_open()) {
        // !!!
        return false;
    }

    if (pimpl->measurement_type_ != Distance) {
        // !!!
        return false;
    }

    data.resize(max_data_size());
    return pimpl->receive_data(&data[0], NULL, time_stamp);
}


bool Lms_driver::get_distance_intensity(std::vector<long>& data,
                                        std::vector<unsigned short>& intensity,
                                        long *time_stamp)
{
    if (!is_open()) {
        // !!!
        return false;
    }

    if (pimpl->measurement_type_ != Distance_intensity) {
        // !!!
        return false;
    }

    size_t data_size = max_data_size();
    data.resize(data_size);
    intensity.resize(data_size);
    return pimpl->receive_data(&data[0], &intensity[0], time_stamp);
}


bool Lms_driver::get_multiecho(std::vector<long>& data_multiecho,
                               long* time_stamp)
{
    if (!is_open()) {
        // !!!
        return false;
    }

    if (pimpl->measurement_type_ != Multiecho) {
        // !!!
        return false;
    }

    size_t data_size = max_data_size() * max_echo_size();
    //fprintf(stderr, "data_size: %ld\n", data_size);
    data_multiecho.resize(data_size);
    return pimpl->receive_data(&data_multiecho[0], NULL, time_stamp);
}


bool Lms_driver::get_multiecho_intensity(std::vector<long>& data_multiecho,
                                         std::vector<unsigned short>&
                                         intensity_multiecho,
                                         long* time_stamp)
{
    if (!is_open()) {
        // !!!
        return false;
    }

    if (pimpl->measurement_type_ != Multiecho_intensity) {
        // !!!
        return false;
    }

    size_t data_size = max_data_size() * max_echo_size();
    //fprintf(stderr, "data_size: %ld\n", data_size);
    data_multiecho.resize(data_size);
    intensity_multiecho.resize(data_size);
    return pimpl->receive_data(&data_multiecho[0],
                               &intensity_multiecho[0], time_stamp);
}


//! \todo 実装する
bool Lms_driver::set_scanning_parameter(int first_step, int last_step,
                                        int skip_step)
{
    (void)first_step;
    (void)last_step;
    (void)skip_step;
    // !!!
    return false;
}


void Lms_driver::stop_measurement()
{
    pimpl->stop_measurement();
}


double Lms_driver::index2rad(int index) const
{
    if (!is_open()) {
        // !!!
        return false;
    }

    if (pimpl->is_multiecho_) {
        index /= max_echo_size();
    }
    return step2rad(index);
}


double Lms_driver::index2deg(int index) const
{
    if (!is_open()) {
        // !!!
        return false;
    }

    return index2rad(index) * 180.0 / M_PI;
}


int Lms_driver::rad2index(double radian) const
{
    if (!is_open()) {
        // !!! set error message
        return -1;
    }
    return pimpl->rad2index(radian);
}


int Lms_driver::deg2index(double degree) const
{
    if (!is_open()) {
        // !!! set error message
        return -1;
    }
    return pimpl->deg2index(degree);
}


double Lms_driver::step2rad(int step) const
{
    if (!is_open()) {
        // !!!
        return false;
    }

    int actual_index = min(max(0, step), pimpl->stop_index_);
    int index = actual_index - pimpl->front_index_ + pimpl->start_index_;
    return (2.0 * M_PI) * index / pimpl->total_steps_;
}


double Lms_driver::step2deg(int step) const
{
    if (!is_open()) {
        // !!!
        return -1;
    }

    return step2rad(step) * 180.0 / M_PI;
}


int Lms_driver::rad2step(double radian) const
{
    if (!is_open()) {
        // !!! set error message
        return -1;
    }

    return pimpl->rad2step(radian);
}


int Lms_driver::deg2step(double degree) const
{
    if (!is_open()) {
        // !!!
        return -1;
    }

    return rad2step(degree * M_PI / 180.0);
}


int Lms_driver::min_step() const
{
    if (!is_open()) {
        // !!!
        return -1;
    }

    return pimpl->min_index_;
}


int Lms_driver::max_step() const
{
    if (!is_open()) {
        // !!!
        return -1;
    }

    return pimpl->max_index_;
}


int Lms_driver::front_step() const
{
    if (!is_open()) {
        // !!!
        return -1;
    }

    return pimpl->front_index_;
}


int Lms_driver::total_steps() const
{
    return pimpl->total_steps_;
}


long Lms_driver::min_distance() const
{
    if (!is_open()) {
        // !!!
        return -1;
    }

    // センサの半径が明らかに 10 [mm] 以上あるので
    return 10;
}


long Lms_driver::max_distance() const
{
    // 距離を表現する hexadecimal の byte 数が 5 より
    return 1 << (4 * 5);
}


long Lms_driver::scan_usec() const
{
    return pimpl->scan_usec_;
}


int Lms_driver::max_data_size() const
{
    if (!is_open()) {
        // !!!
        return -1;
    }
    return pimpl->max_index_ + 1;
}


int Lms_driver::max_echo_size() const
{
    return pimpl->max_echo_size();
}


std::string Lms_driver::sensor_series_name() const
{
    return pimpl->sensor_series_name_;
}


std::string Lms_driver::sensor_product_name() const
{
    return pimpl->sensor_product_name_;
}


std::string Lms_driver::sensor_product_version() const
{
    return pimpl->sensor_product_version_;
}


std::string Lms_driver::serial_id() const
{
    return pimpl->serial_id_;
}
